@@ -15,15 +15,14 @@ from marionette_driver.marionette import Alert
 from marionette_harness import (
     MarionetteTestCase,
     run_if_manage_instance,
-    skip,
     WindowManagerMixin,
 )
 
 here = os.path.abspath(os.path.dirname(__file__))
 
 
-BLACK_PIXEL = 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNkYPhfDwAChwGA60e6kgAAAABJRU5ErkJggg==' # noqa
-RED_PIXEL = 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQMAAAAl21bKAAAAA1BMVEX/TQBcNTh/AAAAAXRSTlPM0jRW/QAAAApJREFUeJxjYgAAAAYAAzY3fKgAAAAASUVORK5CYII=' # noqa
+BLACK_PIXEL = "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNkYPhfDwAChwGA60e6kgAAAABJRU5ErkJggg=="  # noqa
+RED_PIXEL = "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQMAAAAl21bKAAAAA1BMVEX/TQBcNTh/AAAAAXRSTlPM0jRW/QAAAApJREFUeJxjYgAAAAYAAzY3fKgAAAAASUVORK5CYII="  # noqa
 
 
 def inline(doc):
@@ -31,23 +30,23 @@ def inline(doc):
 
 
 def inline_image(data):
-    return 'data:image/png;base64,%s' % data
+    return "data:image/png;base64,%s" % data
 
 
 class BaseNavigationTestCase(WindowManagerMixin, MarionetteTestCase):
-
     def setUp(self):
         super(BaseNavigationTestCase, self).setUp()
 
-        file_path = os.path.join(here, 'data', 'test.html').replace("\\", "/")
+        file_path = os.path.join(here, "data", "test.html").replace("\\", "/")
 
         self.test_page_file_url = "file:///{}".format(file_path)
         self.test_page_frameset = self.marionette.absolute_url("frameset.html")
         self.test_page_insecure = self.fixtures.where_is("test.html", on="https")
         self.test_page_not_remote = "about:robots"
-        self.test_page_push_state = self.marionette.absolute_url("navigation_pushstate.html")
+        self.test_page_push_state = self.marionette.absolute_url(
+            "navigation_pushstate.html"
+        )
         self.test_page_remote = self.marionette.absolute_url("test.html")
-        self.test_page_slow_resource = self.marionette.absolute_url("slow_resource.html")
 
         if self.marionette.session_capabilities["platformName"] == "mac":
             self.mod_key = Keys.META
@@ -59,11 +58,11 @@ class BaseNavigationTestCase(WindowManagerMixin, MarionetteTestCase):
         self.marionette.switch_to_window(self.new_tab)
         Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             lambda _: self.history_length == 1,
-            message="The newly opened tab doesn't have a browser history length of 1")
+            message="The newly opened tab doesn't have a browser history length of 1",
+        )
 
     def tearDown(self):
         self.marionette.timeout.reset()
-        self.marionette.switch_to_parent_frame()
 
         self.close_all_tabs()
 
@@ -77,16 +76,23 @@ class BaseNavigationTestCase(WindowManagerMixin, MarionetteTestCase):
     def is_remote_tab(self):
         with self.marionette.using_context("chrome"):
             # TODO: DO NOT USE MOST RECENT WINDOW BUT CURRENT ONE
-            return self.marionette.execute_script("""
-              Components.utils.import("resource://gre/modules/AppConstants.jsm");
+            return self.marionette.execute_script(
+                """
+              const { AppConstants } = ChromeUtils.import(
+                "resource://gre/modules/AppConstants.jsm"
+              );
 
               let win = null;
 
               if (AppConstants.MOZ_APP_NAME == "fennec") {
-                Components.utils.import("resource://gre/modules/Services.jsm");
+                const { Services } = ChromeUtils.import(
+                  "resource://gre/modules/Services.jsm"
+                );
                 win = Services.wm.getMostRecentWindow("navigator:browser");
               } else {
-                Components.utils.import("resource:///modules/BrowserWindowTracker.jsm");
+                const { BrowserWindowTracker } = ChromeUtils.import(
+                  "resource:///modules/BrowserWindowTracker.jsm"
+                );
                 win = BrowserWindowTracker.getTopWindow();
               }
 
@@ -105,16 +111,17 @@ class BaseNavigationTestCase(WindowManagerMixin, MarionetteTestCase):
               }
 
               return tabBrowser.isRemoteBrowser;
-            """)
+            """
+            )
 
     @property
     def ready_state(self):
-        return self.marionette.execute_script("return window.document.readyState;",
-                                              sandbox=None)
+        return self.marionette.execute_script(
+            "return window.document.readyState;", sandbox=None
+        )
 
 
 class TestNavigate(BaseNavigationTestCase):
-
     def test_set_location_through_execute_script(self):
         # To avoid unexpected remoteness changes and a hang in any non-navigation
         # command (bug 1519354) when navigating via the location bar, already
@@ -123,21 +130,33 @@ class TestNavigate(BaseNavigationTestCase):
 
         self.marionette.execute_script(
             "window.location.href = arguments[0];",
-            script_args=(self.test_page_remote,), sandbox=None)
+            script_args=(self.test_page_remote,),
+            sandbox=None,
+        )
 
         Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             expected.element_present(*(By.ID, "testh1")),
-            message="Target element 'testh1' has not been found")
+            message="Target element 'testh1' has not been found",
+        )
 
         self.assertEqual(self.test_page_remote, self.marionette.get_url())
 
     def test_navigate_chrome_unsupported_error(self):
         with self.marionette.using_context("chrome"):
-            self.assertRaises(errors.UnsupportedOperationException,
-                              self.marionette.navigate, "about:blank")
-            self.assertRaises(errors.UnsupportedOperationException, self.marionette.go_back)
-            self.assertRaises(errors.UnsupportedOperationException, self.marionette.go_forward)
-            self.assertRaises(errors.UnsupportedOperationException, self.marionette.refresh)
+            self.assertRaises(
+                errors.UnsupportedOperationException,
+                self.marionette.navigate,
+                "about:blank",
+            )
+            self.assertRaises(
+                errors.UnsupportedOperationException, self.marionette.go_back
+            )
+            self.assertRaises(
+                errors.UnsupportedOperationException, self.marionette.go_forward
+            )
+            self.assertRaises(
+                errors.UnsupportedOperationException, self.marionette.refresh
+            )
 
     def test_get_current_url_returns_top_level_browsing_context_url(self):
         page_iframe = self.marionette.absolute_url("test_iframe.html")
@@ -158,8 +177,12 @@ class TestNavigate(BaseNavigationTestCase):
         self.marionette.navigate(self.test_page_frameset)
         frame = self.marionette.find_element(By.NAME, "third")
         self.marionette.switch_to_frame(frame)
-        self.assertRaises(errors.NoSuchElementException,
-                          self.marionette.find_element, By.NAME, "third")
+        self.assertRaises(
+            errors.NoSuchElementException,
+            self.marionette.find_element,
+            By.NAME,
+            "third",
+        )
 
         self.marionette.navigate(self.test_page_frameset)
         self.marionette.find_element(By.NAME, "third")
@@ -202,8 +225,9 @@ class TestNavigate(BaseNavigationTestCase):
         self.marionette.navigate(doc)
         self.marionette.execute_script("window.visited = true", sandbox=None)
         self.marionette.navigate("{}#foo".format(doc))
-        self.assertTrue(self.marionette.execute_script(
-            "return window.visited", sandbox=None))
+        self.assertTrue(
+            self.marionette.execute_script("return window.visited", sandbox=None)
+        )
 
     def test_navigate_hash_argument_identical(self):
         test_page = "{}#foo".format(inline("<p id=foo>"))
@@ -270,11 +294,43 @@ class TestNavigate(BaseNavigationTestCase):
         self.marionette.navigate("about:robots")
         self.assertFalse(self.is_remote_tab)
 
+    def test_stale_element_after_remoteness_change(self):
+        self.marionette.navigate(self.test_page_file_url)
+        self.assertTrue(self.is_remote_tab)
+        elem = self.marionette.find_element(By.ID, "file-url")
+
+        self.marionette.navigate("about:robots")
+        self.assertFalse(self.is_remote_tab)
+
+        # Force a GC to get rid of the replaced browsing context.
+        with self.marionette.using_context("chrome"):
+            self.marionette.execute_async_script(
+                """
+                const resolve = arguments[0];
+
+                var memSrv = Cc["@mozilla.org/memory-reporter-manager;1"]
+                  .getService(Ci.nsIMemoryReporterManager);
+
+                Services.obs.notifyObservers(null, "child-mmu-request", null);
+                memSrv.minimizeMemoryUsage(resolve);
+            """
+            )
+
+        with self.assertRaises(errors.StaleElementException):
+            elem.click()
+
     def test_about_blank_for_new_docshell(self):
         self.assertEqual(self.marionette.get_url(), "about:blank")
 
         self.marionette.navigate("about:blank")
 
+    def test_about_newtab(self):
+        with self.marionette.using_prefs({"browser.newtabpage.enabled": True}):
+            self.marionette.navigate("about:newtab")
+
+            self.marionette.navigate(self.test_page_remote)
+            self.marionette.find_element(By.ID, "testDiv")
+
     @run_if_manage_instance("Only runnable if Marionette manages the instance")
     def test_focus_after_navigation(self):
         self.marionette.restart()
@@ -303,7 +359,8 @@ class TestNavigate(BaseNavigationTestCase):
 
         Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             lambda mn: mn.get_url() == "about:support",
-            message="'about:support' hasn't been loaded")
+            message="'about:support' hasn't been loaded",
+        )
         self.assertFalse(self.is_remote_tab)
 
     def test_type_to_remote_tab(self):
@@ -317,12 +374,19 @@ class TestNavigate(BaseNavigationTestCase):
 
         Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             lambda mn: mn.get_url() == self.test_page_remote,
-            message="'{}' hasn't been loaded".format(self.test_page_remote))
+            message="'{}' hasn't been loaded".format(self.test_page_remote),
+        )
         self.assertTrue(self.is_remote_tab)
 
+    def test_navigate_after_deleting_session(self):
+        self.marionette.delete_session()
+        self.marionette.start_session()
+
+        self.marionette.navigate(self.test_page_remote)
+        self.assertEqual(self.test_page_remote, self.marionette.get_url())
+
 
 class TestBackForwardNavigation(BaseNavigationTestCase):
-
     def run_bfcache_test(self, test_pages):
         # Helper method to run simple back and forward testcases.
 
@@ -335,13 +399,17 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
                 else:
                     self.assertEqual(Alert(self.marionette).text, page["alert_text"])
 
-            self.assertEqual(page["url"], self.marionette.get_url())
+            self.assertEqual(self.marionette.get_url(), page["url"])
             self.assertEqual(self.history_length, expected_history_length)
 
             if "is_remote" in page:
-                self.assertEqual(page["is_remote"], self.is_remote_tab,
-                                 "'{}' doesn't match expected remoteness state: {}".format(
-                                     page["url"], page["is_remote"]))
+                self.assertEqual(
+                    page["is_remote"],
+                    self.is_remote_tab,
+                    "'{}' doesn't match expected remoteness state: {}".format(
+                        page["url"], page["is_remote"]
+                    ),
+                )
 
             if "callback" in page and callable(page["callback"]):
                 page["callback"]()
@@ -363,7 +431,6 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
                     self.marionette.go_back()
             else:
                 self.marionette.go_back()
-            self.assertEqual(page["url"], self.marionette.get_url())
 
             check_page_status(page, len(test_pages))
 
@@ -374,7 +441,6 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
                     self.marionette.go_forward()
             else:
                 self.marionette.go_forward()
-            self.assertEqual(page["url"], self.marionette.get_url())
 
             check_page_status(page, len(test_pages))
 
@@ -384,14 +450,16 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
         self.marionette.go_forward()
 
     def test_dismissed_beforeunload_prompt(self):
-        url_beforeunload = inline("""
+        url_beforeunload = inline(
+            """
           <input type="text">
           <script>
             window.addEventListener("beforeunload", function (event) {
               event.preventDefault();
             });
           </script>
-        """)
+        """
+        )
 
         def modify_page():
             self.marionette.find_element(By.TAG_NAME, "input").send_keys("foo")
@@ -404,7 +472,6 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
 
         self.run_bfcache_test(test_pages)
 
-    @skip("Bug 1484927: Maybe causes crash in [@ mozilla::ShutdownXPCOM(nsIServiceManager*)]")
     def test_data_urls(self):
         test_pages = [
             {"url": inline("<p>foobar</p>")},
@@ -457,7 +524,8 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
         # wait for the page load to be complete.
         Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             expected.element_present(*test_element_locator),
-            message="Target element 'email' has not been found")
+            message="Target element 'email' has not been found",
+        )
         self.assertEqual(self.history_length, 3)
 
         # Go back to the frame the click navigated away from
@@ -549,7 +617,8 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
 
         delay = Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             expected.element_present(By.ID, "delay"),
-            message="Target element 'delay' has not been found after timeout in 'back'")
+            message="Target element 'delay' has not been found after timeout in 'back'",
+        )
         self.assertEqual(delay.text, "3")
 
         self.marionette.go_forward()
@@ -563,31 +632,37 @@ class TestBackForwardNavigation(BaseNavigationTestCase):
 
         delay = Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
             expected.element_present(By.ID, "delay"),
-            message="Target element 'delay' has not been found after timeout in 'forward'")
+            message="Target element 'delay' has not been found after timeout in 'forward'",
+        )
         self.assertEqual(delay.text, "4")
 
     def test_certificate_error(self):
         test_pages = [
-            {"url": self.test_page_insecure,
-             "error": errors.InsecureCertificateException},
+            {
+                "url": self.test_page_insecure,
+                "error": errors.InsecureCertificateException,
+            },
             {"url": self.test_page_remote},
-            {"url": self.test_page_insecure,
-             "error": errors.InsecureCertificateException},
+            {
+                "url": self.test_page_insecure,
+                "error": errors.InsecureCertificateException,
+            },
         ]
         self.run_bfcache_test(test_pages)
 
 
 class TestRefresh(BaseNavigationTestCase):
-
     def test_basic(self):
         self.marionette.navigate(self.test_page_remote)
         self.assertEqual(self.test_page_remote, self.marionette.get_url())
 
-        self.marionette.execute_script("""
+        self.marionette.execute_script(
+            """
           let elem = window.document.createElement('div');
           elem.id = 'someDiv';
           window.document.body.appendChild(elem);
-        """)
+        """
+        )
         self.marionette.find_element(By.ID, "someDiv")
 
         self.marionette.refresh()
@@ -601,8 +676,12 @@ class TestRefresh(BaseNavigationTestCase):
 
         frame = self.marionette.find_element(By.NAME, "third")
         self.marionette.switch_to_frame(frame)
-        self.assertRaises(errors.NoSuchElementException,
-                          self.marionette.find_element, By.NAME, "third")
+        self.assertRaises(
+            errors.NoSuchElementException,
+            self.marionette.find_element,
+            By.NAME,
+            "third",
+        )
 
         self.marionette.refresh()
         self.marionette.find_element(By.NAME, "third")
@@ -615,14 +694,18 @@ class TestRefresh(BaseNavigationTestCase):
         self.assertEqual(self.test_page_file_url, self.marionette.get_url())
 
     def test_dismissed_beforeunload_prompt(self):
-        self.marionette.navigate(inline("""
+        self.marionette.navigate(
+            inline(
+                """
           <input type="text">
           <script>
             window.addEventListener("beforeunload", function (event) {
               event.preventDefault();
             });
           </script>
-        """))
+        """
+            )
+        )
         self.marionette.find_element(By.TAG_NAME, "input").send_keys("foo")
         self.marionette.refresh()
 
@@ -631,7 +714,7 @@ class TestRefresh(BaseNavigationTestCase):
             Alert(self.marionette).text
 
     def test_image(self):
-        image = self.marionette.absolute_url('black.png')
+        image = self.marionette.absolute_url("black.png")
 
         self.marionette.navigate(image)
         self.assertEqual(image, self.marionette.get_url())
@@ -673,7 +756,7 @@ class TestRefresh(BaseNavigationTestCase):
     def test_insecure_error(self):
         with self.assertRaises(errors.InsecureCertificateException):
             self.marionette.navigate(self.test_page_insecure)
-        self.assertEqual(self.test_page_insecure, self.marionette.get_url())
+        self.assertEqual(self.marionette.get_url(), self.test_page_insecure)
 
         with self.assertRaises(errors.InsecureCertificateException):
             self.marionette.refresh()
@@ -710,7 +793,8 @@ class TestTLSNavigation(BaseNavigationTestCase):
             self.marionette.switch_to_window(self.new_tab)
             Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
                 lambda _: self.history_length == 1,
-                message="The newly opened tab doesn't have a browser history length of 1")
+                message="The newly opened tab doesn't have a browser history length of 1",
+            )
             yield self.marionette
         finally:
             self.close_all_tabs()
@@ -726,7 +810,8 @@ class TestTLSNavigation(BaseNavigationTestCase):
             self.marionette.switch_to_window(self.new_tab)
             Wait(self.marionette, timeout=self.marionette.timeout.page_load).until(
                 lambda _: self.history_length == 1,
-                message="The newly opened tab doesn't have a browser history length of 1")
+                message="The newly opened tab doesn't have a browser history length of 1",
+            )
             yield self.marionette
         finally:
             self.close_all_tabs()
@@ -739,13 +824,16 @@ class TestTLSNavigation(BaseNavigationTestCase):
     def test_navigate_by_click(self):
         link_url = self.test_page_insecure
         self.marionette.navigate(
-            inline("<a href=%s>https is the future</a>" % link_url))
+            inline("<a href=%s>https is the future</a>" % link_url)
+        )
         self.marionette.find_element(By.TAG_NAME, "a").click()
         self.assertIn("https", self.marionette.get_url())
 
     def test_deactivation(self):
         invalid_cert_url = self.test_page_insecure
 
+        self.marionette.delete_session()
+
         print("with safe session")
         with self.safe_session() as session:
             with self.assertRaises(errors.InsecureCertificateException):
@@ -762,6 +850,22 @@ class TestTLSNavigation(BaseNavigationTestCase):
 
 
 class TestPageLoadStrategy(BaseNavigationTestCase):
+    def setUp(self):
+        super(TestPageLoadStrategy, self).setUp()
+
+        # Test page that delays the response and as such the document to be
+        # loaded. It is used for testing the page load strategy "none".
+        self.test_page_slow = self.marionette.absolute_url("slow")
+
+        # Similar to "slow" but additionally triggers a cross group navigation
+        # which triggers a replacement of the top-level browsing context.
+        self.test_page_slow_coop = self.marionette.absolute_url("slow-coop")
+
+        # Test page that contains a slow loading <img> element which delays the
+        # "load" but not the "DOMContentLoaded" event.
+        self.test_page_slow_resource = self.marionette.absolute_url(
+            "slow_resource.html"
+        )
 
     def tearDown(self):
         self.marionette.delete_session()
@@ -773,17 +877,59 @@ class TestPageLoadStrategy(BaseNavigationTestCase):
         self.marionette.delete_session()
         self.marionette.start_session({"pageLoadStrategy": "none"})
 
-        # With a strategy of "none" there should be no wait for the page load, and the
-        # current load state is unknown. So only test that the command executes successfully.
-        self.marionette.navigate(self.test_page_slow_resource)
+        # Navigate will return immediately. As such wait for the target URL to
+        # be the current location, and the element to exist.
+        self.marionette.navigate(self.test_page_slow)
+        with self.assertRaises(errors.NoSuchElementException):
+            self.marionette.find_element(By.ID, "delay")
+
+        Wait(
+            self.marionette,
+            ignored_exceptions=errors.NoSuchElementException,
+            timeout=self.marionette.timeout.page_load,
+        ).until(lambda _: self.marionette.find_element(By.ID, "delay"))
+
+        self.assertEqual(self.marionette.get_url(), self.test_page_slow)
+
+    def test_none_with_new_session_waits_for_page_loaded(self):
+        self.marionette.delete_session()
+        self.marionette.start_session({"pageLoadStrategy": "none"})
+
+        # Navigate will return immediately.
+        self.marionette.navigate(self.test_page_slow)
+
+        # Make sure that when creating a new session right away it waits
+        # until the page has been finished loading.
+        self.marionette.delete_session()
+        self.marionette.start_session()
+
+        self.assertEqual(self.marionette.get_url(), self.test_page_slow)
+        self.assertEqual(self.ready_state, "complete")
+        self.marionette.find_element(By.ID, "delay")
+
+    def test_none_with_new_session_waits_for_page_loaded_remoteness_change(self):
+        self.marionette.delete_session()
+        self.marionette.start_session({"pageLoadStrategy": "none"})
+
+        # Navigate will return immediately.
+        self.marionette.navigate(self.test_page_slow_coop)
+
+        # Make sure that when creating a new session right away it waits
+        # until the page has been finished loading.
+        self.marionette.delete_session()
+        self.marionette.start_session()
+
+        self.assertEqual(self.marionette.get_url(), self.test_page_slow_coop)
+        self.assertEqual(self.ready_state, "complete")
+        self.marionette.find_element(By.ID, "delay")
 
     def test_eager(self):
         self.marionette.delete_session()
         self.marionette.start_session({"pageLoadStrategy": "eager"})
 
         self.marionette.navigate(self.test_page_slow_resource)
-        self.assertEqual("interactive", self.ready_state)
-        self.assertEqual(self.test_page_slow_resource, self.marionette.get_url())
+        self.assertEqual(self.ready_state, "interactive")
+        self.assertEqual(self.marionette.get_url(), self.test_page_slow_resource)
         self.marionette.find_element(By.ID, "slow")
 
     def test_normal(self):
@@ -791,8 +937,8 @@ class TestPageLoadStrategy(BaseNavigationTestCase):
         self.marionette.start_session({"pageLoadStrategy": "normal"})
 
         self.marionette.navigate(self.test_page_slow_resource)
-        self.assertEqual(self.test_page_slow_resource, self.marionette.get_url())
-        self.assertEqual("complete", self.ready_state)
+        self.assertEqual(self.marionette.get_url(), self.test_page_slow_resource)
+        self.assertEqual(self.ready_state, "complete")
         self.marionette.find_element(By.ID, "slow")
 
     def test_strategy_after_remoteness_change(self):
@@ -801,8 +947,10 @@ class TestPageLoadStrategy(BaseNavigationTestCase):
         self.marionette.start_session({"pageLoadStrategy": "eager"})
 
         # Trigger a remoteness change which will reload the listener script
-        self.assertTrue(self.is_remote_tab, "Initial tab doesn't have remoteness flag set")
+        self.assertTrue(
+            self.is_remote_tab, "Initial tab doesn't have remoteness flag set"
+        )
         self.marionette.navigate("about:robots")
         self.assertFalse(self.is_remote_tab, "Tab has remoteness flag set")
         self.marionette.navigate(self.test_page_slow_resource)
-        self.assertEqual("interactive", self.ready_state)
+        self.assertEqual(self.ready_state, "interactive")
