@@ -6,13 +6,13 @@
 
 /*
  * This file contains common code that is loaded before each test file(s).
- * See http://developer.mozilla.org/en/docs/Writing_xpcshell-based_unit_tests
+ * See https://developer.mozilla.org/en-US/docs/Mozilla/QA/Writing_xpcshell-based_unit_tests
  * for more information.
  */
 
 /* defined by the harness */
 /* globals _HEAD_FILES, _HEAD_JS_PATH, _JSDEBUGGER_PORT, _JSCOV_DIR,
-    _MOZINFO_JS_PATH, _TEST_FILE, _TEST_NAME, _TESTING_MODULES_DIR:true,
+    _MOZINFO_JS_PATH, _TEST_FILE, _TEST_NAME, _TEST_CWD, _TESTING_MODULES_DIR:true,
     _PREFS_FILE */
 
 /* defined by XPCShellImpl.cpp */
@@ -41,32 +41,34 @@ var _XPCSHELL_PROCESS;
 
 // Register the testing-common resource protocol early, to have access to its
 // modules.
-var _Services = ChromeUtils.import("resource://gre/modules/Services.jsm", null)
-  .Services;
+let { Services: _Services } = ChromeUtils.import(
+  "resource://gre/modules/Services.jsm"
+);
 _register_modules_protocol_handler();
 
-var _AppConstants = ChromeUtils.import(
-  "resource://gre/modules/AppConstants.jsm",
-  null
-).AppConstants;
+let { AppConstants: _AppConstants } = ChromeUtils.import(
+  "resource://gre/modules/AppConstants.jsm"
+);
 
-var _PromiseTestUtils = ChromeUtils.import(
-  "resource://testing-common/PromiseTestUtils.jsm",
-  null
-).PromiseTestUtils;
-var _Task = ChromeUtils.import("resource://testing-common/Task.jsm", null).Task;
+let { PromiseTestUtils: _PromiseTestUtils } = ChromeUtils.import(
+  "resource://testing-common/PromiseTestUtils.jsm"
+);
 
-let _NetUtil = ChromeUtils.import("resource://gre/modules/NetUtil.jsm", null)
-  .NetUtil;
+let { NetUtil: _NetUtil } = ChromeUtils.import(
+  "resource://gre/modules/NetUtil.jsm"
+);
 
-let _XPCOMUtils = ChromeUtils.import(
-  "resource://gre/modules/XPCOMUtils.jsm",
-  null
-).XPCOMUtils;
+let { XPCOMUtils: _XPCOMUtils } = ChromeUtils.import(
+  "resource://gre/modules/XPCOMUtils.jsm"
+);
+
+let { OS: _OS } = ChromeUtils.import("resource://gre/modules/osfile.jsm");
 
 // Support a common assertion library, Assert.jsm.
-var AssertCls = ChromeUtils.import("resource://testing-common/Assert.jsm", null)
-  .Assert;
+var { Assert: AssertCls } = ChromeUtils.import(
+  "resource://testing-common/Assert.jsm"
+);
+
 // Pass a custom report function for xpcshell-test style reporting.
 var Assert = new AssertCls(function(err, message, stack) {
   if (err) {
@@ -92,10 +94,9 @@ var _dumpLog = function(raw_msg) {
   dump("\n" + JSON.stringify(raw_msg) + "\n");
 };
 
-var _LoggerClass = ChromeUtils.import(
-  "resource://testing-common/StructuredLog.jsm",
-  null
-).StructuredLogger;
+var { StructuredLogger: _LoggerClass } = ChromeUtils.import(
+  "resource://testing-common/StructuredLog.jsm"
+);
 var _testLogger = new _LoggerClass("xpcshell/head.js", _dumpLog, [_add_params]);
 
 // Disable automatic network detection, so tests work correctly when
@@ -121,15 +122,6 @@ if (runningInParent && "mozIAsyncHistory" in Ci) {
   _Services.prefs.setBoolPref("places.history.enabled", true);
 }
 
-try {
-  if (runningInParent) {
-    // Disable IPv6 lookups for 'localhost' on windows.
-    if ("@mozilla.org/windows-registry-key;1" in Cc) {
-      _Services.prefs.setCharPref("network.dns.ipv4OnlyDomains", "localhost");
-    }
-  }
-} catch (e) {}
-
 // Configure crash reporting, if possible
 // We rely on the Python harness to set MOZ_CRASHREPORTER,
 // MOZ_CRASHREPORTER_NO_REPORT, and handle checking for minidumps.
@@ -245,7 +237,7 @@ function _do_main() {
 
   var tm = Cc["@mozilla.org/thread-manager;1"].getService();
 
-  tm.spinEventLoopUntil(() => _quit);
+  tm.spinEventLoopUntil("Test(xpcshell/head.js:_do_main)", () => _quit);
 
   tm.spinEventLoopUntilEmpty();
 }
@@ -277,7 +269,7 @@ var _fakeIdleService = {
       Ci.nsIComponentRegistrar
     ));
   },
-  contractID: "@mozilla.org/widget/idleservice;1",
+  contractID: "@mozilla.org/widget/useridleservice;1",
   CID: Components.ID("{9163a4ae-70c2-446c-9ac1-bbe4ab93004e}"),
 
   activate: function FIS_activate() {
@@ -310,19 +302,13 @@ var _fakeIdleService = {
 
   factory: {
     // nsIFactory
-    createInstance(aOuter, aIID) {
-      if (aOuter) {
-        throw Components.Exception("", Cr.NS_ERROR_NO_AGGREGATION);
-      }
+    createInstance(aIID) {
       return _fakeIdleService.QueryInterface(aIID);
     },
-    lockFactory(aLock) {
-      throw Components.Exception("", Cr.NS_ERROR_NOT_IMPLEMENTED);
-    },
     QueryInterface: ChromeUtils.generateQI(["nsIFactory"]),
   },
 
-  // nsIIdleService
+  // nsIUserIdleService
   get idleTime() {
     return 0;
   },
@@ -335,7 +321,7 @@ var _fakeIdleService = {
     if (aIID.equals(Ci.nsIFactory)) {
       return this.factory;
     }
-    if (aIID.equals(Ci.nsIIdleService) || aIID.equals(Ci.nsISupports)) {
+    if (aIID.equals(Ci.nsIUserIdleService) || aIID.equals(Ci.nsISupports)) {
       return this;
     }
     throw Components.Exception("", Cr.NS_ERROR_NO_INTERFACE);
@@ -348,7 +334,7 @@ var _fakeIdleService = {
  */
 function do_get_idle() {
   _fakeIdleService.deactivate();
-  return Cc[_fakeIdleService.contractID].getService(Ci.nsIIdleService);
+  return Cc[_fakeIdleService.contractID].getService(Ci.nsIUserIdleService);
 }
 
 // Map resource://test/ to current working directory and
@@ -416,7 +402,9 @@ function _setupDevToolsServer(breakpointFiles, callback) {
 
   let require;
   try {
-    ({ require } = ChromeUtils.import("resource://devtools/shared/Loader.jsm"));
+    ({ require } = ChromeUtils.import(
+      "resource://devtools/shared/loader/Loader.jsm"
+    ));
   } catch (e) {
     throw new Error(
       "resource://devtools appears to be inaccessible from the " +
@@ -437,18 +425,17 @@ function _setupDevToolsServer(breakpointFiles, callback) {
   DevToolsServer.setRootActor(createRootActor);
   DevToolsServer.allowChromeProcess = true;
 
-  // An observer notification that tells us when we can "resume" script
-  // execution.
   const TOPICS = [
-    "devtools-thread-instantiated",
-    "devtools-thread-resumed",
+    // An observer notification that tells us when the thread actor is ready
+    // and can accept breakpoints.
+    "devtools-thread-ready",
+    // Or when devtools are destroyed and we should stop observing.
     "xpcshell-test-devtools-shutdown",
   ];
   let observe = function(subject, topic, data) {
-    if (topic === "devtools-thread-instantiated") {
+    if (topic === "devtools-thread-ready") {
       const threadActor = subject.wrappedJSObject;
       threadActor.setBreakpointOnLoad(breakpointFiles);
-      return;
     }
 
     for (let topicToRemove of TOPICS) {
@@ -499,7 +486,7 @@ function _initDebugging(port) {
 
   // spin an event loop until the debugger connects.
   const tm = Cc["@mozilla.org/thread-manager;1"].getService();
-  tm.spinEventLoopUntil(() => {
+  tm.spinEventLoopUntil("Test(xpcshell/head.js:_initDebugging)", () => {
     if (initialized) {
       return true;
     }
@@ -513,6 +500,40 @@ function _initDebugging(port) {
 }
 
 function _execute_test() {
+  if (typeof _TEST_CWD != "undefined") {
+    let cwd_complete = false;
+    _OS.File.setCurrentDirectory(_TEST_CWD)
+      .then(_ => (cwd_complete = true))
+      .catch(e => {
+        _testLogger.error(_exception_message(e));
+        cwd_complete = true;
+      });
+    _Services.tm.spinEventLoopUntil(
+      "Test(xpcshell/head.js:setCurrentDirectory)",
+      () => cwd_complete
+    );
+  }
+
+  if (runningInParent && _AppConstants.platform == "android") {
+    try {
+      // GeckoView initialization needs the profile
+      do_get_profile(true);
+      // Wake up GeckoViewStartup
+      let geckoViewStartup = Cc["@mozilla.org/geckoview/startup;1"].getService(
+        Ci.nsIObserver
+      );
+      geckoViewStartup.observe(null, "profile-after-change", null);
+      geckoViewStartup.observe(null, "app-startup", null);
+
+      // Glean needs to be initialized for metric recording & tests to work.
+      // Usually this happens through Glean Kotlin,
+      // but for xpcshell tests we initialize it from here.
+      _Services.fog.initializeFOG();
+    } catch (ex) {
+      do_throw(`Failed to initialize GeckoView: ${ex}`, ex.stack);
+    }
+  }
+
   // _JSDEBUGGER_PORT is dynamically defined by <runxpcshelltests.py>.
   if (_JSDEBUGGER_PORT) {
     try {
@@ -535,8 +556,7 @@ function _execute_test() {
   let coverageCollector = null;
   if (typeof _JSCOV_DIR === "string") {
     let _CoverageCollector = ChromeUtils.import(
-      "resource://testing-common/CoverageUtils.jsm",
-      {}
+      "resource://testing-common/CoverageUtils.jsm"
     ).CoverageCollector;
     coverageCollector = new _CoverageCollector(_JSCOV_DIR);
   }
@@ -650,16 +670,23 @@ function _execute_test() {
   })()
     .catch(reportCleanupError)
     .then(() => (complete = true));
-  _Services.tm.spinEventLoopUntil(() => complete);
+  _Services.tm.spinEventLoopUntil(
+    "Test(xpcshell/head.js:_execute_test)",
+    () => complete
+  );
   if (cleanupStartTime) {
     ChromeUtils.addProfilerMarker(
       "xpcshell-test",
-      cleanupStartTime,
+      { category: "Test", startTime: cleanupStartTime },
       "Cleanup functions"
     );
   }
 
-  ChromeUtils.addProfilerMarker("xpcshell-test", startTime, _TEST_NAME);
+  ChromeUtils.addProfilerMarker(
+    "xpcshell-test",
+    { category: "Test", startTime },
+    _TEST_NAME
+  );
   _Services.obs.notifyObservers(null, "test-complete");
 
   // Restore idle service to avoid leaks.
@@ -668,10 +695,18 @@ function _execute_test() {
   if (_profileInitialized) {
     // Since we have a profile, we will notify profile shutdown topics at
     // the end of the current test, to ensure correct cleanup on shutdown.
-    _Services.obs.notifyObservers(null, "profile-change-net-teardown");
-    _Services.obs.notifyObservers(null, "profile-change-teardown");
-    _Services.obs.notifyObservers(null, "profile-before-change");
-    _Services.obs.notifyObservers(null, "profile-before-change-qm");
+    _Services.startup.advanceShutdownPhase(
+      _Services.startup.SHUTDOWN_PHASE_APPSHUTDOWNNETTEARDOWN
+    );
+    _Services.startup.advanceShutdownPhase(
+      _Services.startup.SHUTDOWN_PHASE_APPSHUTDOWNTEARDOWN
+    );
+    _Services.startup.advanceShutdownPhase(
+      _Services.startup.SHUTDOWN_PHASE_APPSHUTDOWN
+    );
+    _Services.startup.advanceShutdownPhase(
+      _Services.startup.SHUTDOWN_PHASE_APPSHUTDOWNQM
+    );
 
     _profileInitialized = false;
   }
@@ -719,9 +754,9 @@ function _load_files(aFiles) {
       let startTime = Cu.now();
       load(element);
       ChromeUtils.addProfilerMarker(
-        "xpcshell-test",
-        startTime,
-        "load " + element.replace(/.*\/_?tests\/xpcshell\//, "")
+        "load_file",
+        { category: "Test", startTime },
+        element.replace(/.*\/_?tests\/xpcshell\//, "")
       );
     } catch (e) {
       let extra = {
@@ -747,7 +782,7 @@ function _wrap_with_quotes_if_necessary(val) {
  * Prints a message to the output log.
  */
 function info(msg, data) {
-  ChromeUtils.addProfilerMarker("xpcshell-test", undefined, "INFO " + msg);
+  ChromeUtils.addProfilerMarker("INFO", { category: "Test" }, msg);
   msg = _wrap_with_quotes_if_necessary(msg);
   data = data ? data : null;
   _testLogger.info(msg, data);
@@ -844,7 +879,7 @@ function _format_stack(stack) {
   } else {
     normalized = "" + stack;
   }
-  return _Task.Debugging.generateReadableStack(normalized, "    ");
+  return normalized;
 }
 
 // Make a nice display string from an object that behaves
@@ -1361,6 +1396,10 @@ function do_load_child_test_harness() {
     command += " const _JSCOV_DIR=" + uneval(_JSCOV_DIR) + ";";
   }
 
+  if (typeof _TEST_CWD != "undefined") {
+    command += " const _TEST_CWD=" + uneval(_TEST_CWD) + ";";
+  }
+
   if (_TESTING_MODULES_DIR) {
     command +=
       " const _TESTING_MODULES_DIR=" + uneval(_TESTING_MODULES_DIR) + ";";
@@ -1463,6 +1502,18 @@ function do_send_remote_message(name, data) {
   mm[sender](name, data);
 }
 
+/**
+ * Schedules and awaits a precise GC, and forces CC, `maxCount` number of times.
+ * @param maxCount
+ *        How many times GC and CC should be scheduled.
+ */
+async function schedulePreciseGCAndForceCC(maxCount) {
+  for (let count = 0; count < maxCount; count++) {
+    await new Promise(resolve => Cu.schedulePreciseGC(resolve));
+    Cu.forceCC();
+  }
+}
+
 /**
  * Add a test function to the list of tests that are to be run asynchronously.
  *
@@ -1471,10 +1522,14 @@ function do_send_remote_message(name, data) {
  *        Supported properties:
  *          skip_if : An arrow function which has an expression to be
  *                    evaluated whether the test is skipped or not.
+ *          pref_set: An array of preferences to set for the test, reset at end of test.
  * @param func
  *        A function to be run only if the funcOrProperies is not a function.
  * @param isTask
  *        Optional flag that indicates whether `func` is a task. Defaults to `false`.
+ * @param isSetup
+ *        Optional flag that indicates whether `func` is a setup task. Defaults to `false`.
+ *        Implies isTask.
  *
  * Each test function must call run_next_test() when it's done. Test files
  * should call run_next_test() in their run_test function to execute all
@@ -1482,14 +1537,30 @@ function do_send_remote_message(name, data) {
  *
  * @return the test function that was passed in.
  */
+var _gSupportedProperties = ["skip_if", "pref_set"];
 var _gTests = [];
 var _gRunOnlyThisTest = null;
-function add_test(properties, func = properties, isTask = false) {
+function add_test(
+  properties,
+  func = properties,
+  isTask = false,
+  isSetup = false
+) {
+  if (isSetup) {
+    isTask = true;
+  }
   if (typeof properties == "function") {
-    properties = { isTask };
+    properties = { isTask, isSetup };
     _gTests.push([properties, func]);
   } else if (typeof properties == "object") {
+    // Ensure only documented properties are in the object.
+    for (let prop of Object.keys(properties)) {
+      if (!_gSupportedProperties.includes(prop)) {
+        do_throw(`Task property is not supported: ${prop}`);
+      }
+    }
     properties.isTask = isTask;
+    properties.isSetup = isSetup;
     _gTests.push([properties, func]);
   } else {
     do_throw("add_test() should take a function or an object and a function");
@@ -1507,6 +1578,7 @@ function add_test(properties, func = properties, isTask = false) {
  *        Supported properties:
  *          skip_if : An arrow function which has an expression to be
  *                    evaluated whether the test is skipped or not.
+ *          pref_set: An array of preferences to set for the test, reset at end of test.
  * @param func
  *        An async function to be run only if the funcOrProperies is not a function.
  *
@@ -1519,8 +1591,7 @@ function add_test(properties, func = properties, isTask = false) {
  *
  * Unlike add_test(), there is no need to call run_next_test(). The next test
  * will run automatically as soon the task function is exhausted. To trigger
- * premature (but successful) termination of the function, simply return or
- * throw a Task.Result instance.
+ * premature (but successful) termination of the function or simply return.
  *
  * Example usage:
  *
@@ -1546,6 +1617,7 @@ function add_test(properties, func = properties, isTask = false) {
  *
  * add_task({
  *   skip_if: () => !("@mozilla.org/telephony/volume-service;1" in Components.classes),
+ *   pref_set: [["some.pref", "value"], ["another.pref", true]],
  * }, async function test_volume_service() {
  *   let volumeService = Cc["@mozilla.org/telephony/volume-service;1"]
  *     .getService(Ci.nsIVolumeService);
@@ -1556,7 +1628,58 @@ function add_task(properties, func = properties) {
   return add_test(properties, func, true);
 }
 
-_Task.Debugging.maintainStack = true;
+/**
+ * add_setup is like add_task, but creates setup tasks.
+ */
+function add_setup(properties, func = properties) {
+  return add_test(properties, func, true, true);
+}
+
+const _setTaskPrefs = prefs => {
+  for (let [pref, value] of prefs) {
+    if (value === undefined) {
+      // Clear any pref that didn't have a user value.
+      info(`Clearing pref "${pref}"`);
+      _Services.prefs.clearUserPref(pref);
+      continue;
+    }
+
+    info(`Setting pref "${pref}": ${value}`);
+    switch (typeof value) {
+      case "boolean":
+        _Services.prefs.setBoolPref(pref, value);
+        break;
+      case "number":
+        _Services.prefs.setIntPref(pref, value);
+        break;
+      case "string":
+        _Services.prefs.setStringPref(pref, value);
+        break;
+      default:
+        throw new Error("runWithPrefs doesn't support this pref type yet");
+    }
+  }
+};
+
+const _getTaskPrefs = prefs => {
+  return prefs.map(([pref, value]) => {
+    info(`Getting initial pref value for "${pref}"`);
+    if (!_Services.prefs.prefHasUserValue(pref)) {
+      // Check if the pref doesn't have a user value.
+      return [pref, undefined];
+    }
+    switch (typeof value) {
+      case "boolean":
+        return [pref, _Services.prefs.getBoolPref(pref)];
+      case "number":
+        return [pref, _Services.prefs.getIntPref(pref)];
+      case "string":
+        return [pref, _Services.prefs.getStringPref(pref)];
+      default:
+        throw new Error("runWithPrefs doesn't support this pref type yet");
+    }
+  });
+};
 
 /**
  * Runs the next test function from the list of async tests.
@@ -1564,11 +1687,20 @@ _Task.Debugging.maintainStack = true;
 var _gRunningTest = null;
 var _gTestIndex = 0; // The index of the currently running test.
 var _gTaskRunning = false;
+var _gSetupRunning = false;
 function run_next_test() {
   if (_gTaskRunning) {
     throw new Error(
       "run_next_test() called from an add_task() test function. " +
-        "run_next_test() should not be called from inside add_task() " +
+        "run_next_test() should not be called from inside add_setup() or add_task() " +
+        "under any circumstances!"
+    );
+  }
+
+  if (_gSetupRunning) {
+    throw new Error(
+      "run_next_test() called from an add_setup() test function. " +
+        "run_next_test() should not be called from inside add_setup() or add_task() " +
         "under any circumstances!"
     );
   }
@@ -1582,12 +1714,18 @@ function run_next_test() {
 
       // Must set to pending before we check for skip, so that we keep the
       // running counts correct.
-      _testLogger.info(_TEST_NAME + " | Starting " + _gRunningTest.name);
+      _testLogger.info(
+        `${_TEST_NAME} | Starting ${_properties.isSetup ? "setup " : ""}${
+          _gRunningTest.name
+        }`
+      );
       do_test_pending(_gRunningTest.name);
 
       if (
         (typeof _properties.skip_if == "function" && _properties.skip_if()) ||
-        (_gRunOnlyThisTest && _gRunningTest != _gRunOnlyThisTest)
+        (_gRunOnlyThisTest &&
+          _gRunningTest != _gRunOnlyThisTest &&
+          !_properties.isSetup)
       ) {
         let _condition = _gRunOnlyThisTest
           ? "only one task may run."
@@ -1612,29 +1750,41 @@ function run_next_test() {
         return;
       }
 
+      let initialPrefsValues = [];
+      if (_properties.pref_set) {
+        initialPrefsValues = _getTaskPrefs(_properties.pref_set);
+        _setTaskPrefs(_properties.pref_set);
+      }
+
       if (_properties.isTask) {
-        _gTaskRunning = true;
+        if (_properties.isSetup) {
+          _gSetupRunning = true;
+        } else {
+          _gTaskRunning = true;
+        }
         let startTime = Cu.now();
         (async () => _gRunningTest())().then(
           result => {
-            _gTaskRunning = false;
+            _gTaskRunning = _gSetupRunning = false;
             ChromeUtils.addProfilerMarker(
-              "xpcshell-test",
-              startTime,
-              _gRunningTest.name || "task"
+              "task",
+              { category: "Test", startTime },
+              _gRunningTest.name || undefined
             );
             if (_isGenerator(result)) {
               Assert.ok(false, "Task returned a generator");
             }
+            _setTaskPrefs(initialPrefsValues);
             run_next_test();
           },
           ex => {
-            _gTaskRunning = false;
+            _gTaskRunning = _gSetupRunning = false;
             ChromeUtils.addProfilerMarker(
-              "xpcshell-test",
-              startTime,
-              _gRunningTest.name || "task"
+              "task",
+              { category: "Test", startTime },
+              _gRunningTest.name || undefined
             );
+            _setTaskPrefs(initialPrefsValues);
             try {
               do_report_unexpected_exception(ex);
             } catch (error) {
@@ -1653,14 +1803,25 @@ function run_next_test() {
         } finally {
           ChromeUtils.addProfilerMarker(
             "xpcshell-test",
-            startTime,
+            { category: "Test", startTime },
             _gRunningTest.name || undefined
           );
+          _setTaskPrefs(initialPrefsValues);
         }
       }
     }
   }
 
+  function frontLoadSetups() {
+    _gTests.sort(([propsA, funcA], [propsB, funcB]) => {
+      return propsB.isSetup ? 1 : 0;
+    });
+  }
+
+  if (!_gTestIndex) {
+    frontLoadSetups();
+  }
+
   // For sane stacks during failures, we execute this code soon, but not now.
   // We do this now, before we call do_test_finished(), to ensure the pending
   // counter (_tests_pending) never reaches 0 while we still have tests to run
@@ -1696,21 +1857,23 @@ try {
   // We only want to run this for local developer builds (which should have a "default" update channel).
   if (runningInParent && _AppConstants.MOZ_UPDATE_CHANNEL == "default") {
     let startTime = Cu.now();
-    let _TelemetryController = ChromeUtils.import(
-      "resource://gre/modules/TelemetryController.jsm",
-      null
-    ).TelemetryController;
+    let { TelemetryController: _TelemetryController } = ChromeUtils.import(
+      "resource://gre/modules/TelemetryController.jsm"
+    );
 
     let complete = false;
     _TelemetryController.testRegisterJsProbes().finally(() => {
       ChromeUtils.addProfilerMarker(
         "xpcshell-test",
-        startTime,
+        { category: "Test", startTime },
         "TelemetryController.testRegisterJsProbes"
       );
       complete = true;
     });
-    _Services.tm.spinEventLoopUntil(() => complete);
+    _Services.tm.spinEventLoopUntil(
+      "Test(xpcshell/head.js:run_next-Test)",
+      () => complete
+    );
   }
 } catch (e) {
   do_throw(e);
